#include <jstring.h>
#include <vector>
#include <list>
#include "drawgraph.h"
#include "elttypes.h"
#include <jiostr.h>
#include <jstrstr.h>
#include <algorithm>

// ---------------------- Global variables

Coord epsilon=.001; // used to make things slightly different in Z
Settings setting;

// ---------------------- DrawElt descendants

void Line :: Render(DrawTo *dt)
  { dt->DrawLine(P1, P2, R, erase); }

void Disc :: Render(DrawTo *dt)
  { dt->DrawDisc(C, R, erase); }

void Text :: Render(DrawTo *dt)
  { dt->DrawText(ll, text); }

// ---------------------- Settings member functions

Settings :: Settings()
  { DiscRThickSet=LineRThickSet=false; 
    LineRNormal=.375; LineRThick=.9; 
    DiscRNormal=1.5;  DiscRThick=1.875; 
    DrawBaseLabels=true; DrawCoverLabels=true;
    LoopSize=20;  LoopAspect=.5;  MultiEdgeWidth=100;
    detail=10;      // must be at least 1, though 3 is probably a practical minimum
    yisup=true;
    CoverYStretch=1.;
    CoverRandomDisplace=5.0;
    CoverEnableCurves=true;
  }

// returns true if setting is recognized
// assumes name is in lowercase
bool Settings :: Setting(string name, Coord value)
  {
    // Still have to do:
    // scale         1.0
    // downarrow     1

    if (name=="detail")
      {
        if (value >= 1)
          detail=static_cast<int>(value);
        else
          cerr << "Invalid detail value of " << value << " ignored.\n";
        return true;
      }
    else if (name=="yisup")
      {
        yisup=(value >= .5);
        return true;
      }
    else if (name=="linerad")
      {
        if (value>=0.)
          {
            LineRNormal=value;
            if (! LineRThickSet)
              LineRThick=LineRNormal*5./3.;
          }
        else
          cerr << "Illegal linerad value of " << value << " ignored\n";
        return true;
      }
    else if (name=="linehalorad")
      {
        if (value>=0.)
          {
            LineRThick=value;
            LineRThickSet=true;
          }
        else
          cerr << "Illegal linehalorad value of " << value << " ignored\n";
        return true;
      }
    else if (name=="vertexrad")
      {
        if (value>=0.)
          {
            DiscRNormal=value;
            if (! DiscRThickSet)
              DiscRThick=DiscRNormal*5./4.;
          }
        else
          cerr << "Illegal vertexrad value of " << value << " ignored\n";
        return true;
      }
    else if (name=="vertexhalorad")
      {
        if (value>=0.)
          {
            DiscRThick=value;
            DiscRThickSet=true;
          }
        else
          cerr << "Illegal dischalorad value of " << value << " ignored\n";
        return true;
      }
    else if (name=="baselabels")
      {
        DrawBaseLabels=(value>=0.5);
        return true;
      }
    else if (name=="coverlabels")
      {
        DrawCoverLabels=(value>=0.5);
        return true;
      }
    else if (name=="loopsize")
      {
        if (value>=0.)
          LoopSize=value;
        else
          cerr << "Illegal loopsize value of " << value << " ignored.\n";
        return true;
      }
    else if (name=="loopaspect")
      {
        if (value>=0.)
          LoopAspect=value;
        else
          cerr << "Illegal loopaspect value of " << value << " ignored.\n";
        return true;
      }
    else if (name=="multiedgesize")
      {
        if (value>=0.)
          MultiEdgeWidth=value;
        else
          cerr << "Illegal multiedgesize value of " << value << " ignored.\n";
        return true;
      }
    else if (name=="coverystretch")
      {
        if (value>=0.)
          CoverYStretch=value;
        else
          cerr << "Illegal coverystretchvalue of " << value << " ignored.\n";
        return true;
      }
    else if (name=="coverrandomdisplace")
      {
        if (value>=0.)
          CoverRandomDisplace=value;
        else
          cerr << "Illegal coverrandomdisplace of " << value << " ignored.\n";
        return true;
      }
    else if (name=="coverenablecurves")
      {
        CoverEnableCurves=(value>=0.5);
        return true;
      }
    return false;
  }

// ---------------------- Scene member functions

void Scene :: Render() // renders all DrawElt's in order of decreasing Z
  { 
    instl sort(allrenderable.begin(), allrenderable.end(), compare);
    for (iter i=allrenderable.begin(); i!=allrenderable.end(); ++i)
      (*i)->Render(renderto); 
  }

// ---------------------- User Interface functions

void HelpText()
  {
    cerr << "Command line: graphcover [settingname=value ...] < infile > outfile\n"
            "Input file format:\n"
            "    setting:     s settingname value\n"
            "    vertex:      v name x y [bringforwardamount]\n"
            "                   (72 units = 1 inch)\n"
            "    edge:        e name1 name2 [multiplicity]\n"
            "      (where name1 and name2 are vertex names)\n"
            "      Please list the vertices before the edges.\n"
            "If have a cover, then:\n"
            "    begin cover: c\n"
            "    vertex:      v name (should contain the name of a base vertex)\n"
            "    edge:        e name1 name2 [multiplicity]\n"
            "      (where name1 and name2 are names of vertices in the cover)\n"
            "      Again list vertices before edges.\n" 
            "Settings: <name>        <default value>    <name>        <default>\n"
            "          textheight    12.0               coverystretch 1.0\n"
            "          loopsize      20.0               coverrandomdisplace 5.0\n"
            "          loopaspect    0.5                coverenablecurves   1\n"
            "          multiedgesize 100.0              detail        10\n"
            "          linerad       0.375              baselabels    1\n"
            "          linehalorad   0.75               coverlabels   1\n"
            "          vertexrad     1.5                yisup         1\n"
            "          vertexhalorad 1.875              eps           0" << endl;
// "scale         1.0"
// "downarrow     1"
  }

void ParseSetting(DrawTo &drawto, string name, Coord value)
  {
    // lowercase the string
    for (int i=0; i<name.length(); ++i)
      name[i]=tolower(name[i]);

    if (! drawto.Setting(name, value))
      if (! setting.Setting(name, value))
        cerr << "Ignoring unrecognized setting '" << name << "'.\n";
  }

bool ParseInput(DrawTo &drawto, GraphBase &graph, GraphCover &cover)
  {
    bool needhelp=false;
    bool incover=false;
    bool someoutput=false;
    while (cin && ! cin.eof())
      { // process input file
        string s;
        if (cin >> s)
          {
            if (s.length() && isalpha(s[0])) // else a comment
              {
                switch (s[0])
                  {
                    case 's':
                      {
                        instl getline(cin, s);
                        instl istringstream istr(WORKAROUND_OLDISTRSTREAM(s));

                        string name;
                        if (! (istr >> name))
                          {
                            cerr << "Error reading setting name\n";
                            break;
                          }
                        double value;
                        if (! (istr >> value))
                          {
                            cerr << "Error reading setting '" << name << "' value\n";
                            break;
                          }
                        ParseSetting(drawto, name, value);
                      }
                      break;

                    case 'v':
                      {
                        instl getline(cin, s);
                        instl istringstream istr(WORKAROUND_OLDISTRSTREAM(s));
                        string name;
                        if (! (istr >> name))  // get vertex name
                          {
                            cerr << "Error reading vertex\n";
                            break;
                          }

                        if (incover)
                          {
                            int ret=cover.AddVertex(name);
                            if (ret==-1)
                              cerr << "Invalid cover vertex\n";
                            else
                              someoutput=true;
                          }
                        else
                          {
                            Coord p[3];         // get vertex location
                            if (! (istr >> p[0] >> p[2]))
                              {
                                cerr << "Error reading vertex '" << name << "' coordinates\n";
                                break;
                              }
                            if (! setting.yisup)
                              p[2]= -p[2];
                                                // check for optional 'y'
                            if (! (istr >> p[1]))
                              p[1]=0;
                            else
                              p[2]-=p[1];       // subtract y from y+z to get z
                            graph.AddVertex(p, name);
                            someoutput=true;
                          }
                      }
                      break;

                    case 'e':
                      {
                        instl getline(cin, s);
                        instl istringstream istr(WORKAROUND_OLDISTRSTREAM(s));
                        string name;
                        int ret;
                        if (incover)
                          {
                            if (! (istr >> name)) // get vertex 1 name
                              {
                                cerr << "Error reading vertex 1 in edge\n";
                                break;
                              }
                            int v1=cover.FindVertex(name); // find vertex number
                            if (v1==-1)
                              {
                                cerr << "Did not find first vertex '" << name << "' in edge\n"
                                        "  Check spelling and make sure you declared the vertices first.\n";
                                // list verticies?
                                break;
                              }

                            if (! (istr >> name)) // get vertex 2 name
                              {
                                cerr << "Error reading vertex 2 in edge\n";
                                break;
                              }
                            int v2=cover.FindVertex(name); // find vertex number
                            if (v2==-1)
                              {
                                cerr << "Did not find second vertex '" << name << "' in edge\n"
                                        "  Check spelling and make sure you declared the vertices first.\n";
                                // list verticies?
                                break;
                              }
                            int mult;
                            if (! (istr >> mult)) // get multiplicity
                              mult=1;
                            ret=cover.AddEdge(v1, v2, mult);
                          }
                        else
                          {
                            if (! (istr >> name)) // get vertex 1 name
                              {
                                cerr << "Error reading vertex 1 in edge\n";
                                break;
                              }
                            int v1=graph.FindVertex(name); // find vertex number
                            if (v1==-1)
                              {
                                cerr << "Did not find first vertex '" << name << "' in edge\n"
                                        "  Check spelling and make sure you declared the vertices first.\n";
                                // list verticies?
                                break;
                              }

                            if (! (istr >> name)) // get vertex 2 name
                              {
                                cerr << "Error reading vertex 2 in edge\n";
                                break;
                              }
                            int v2=graph.FindVertex(name); // find vertex number
                            if (v2==-1)
                              {
                                cerr << "Did not find second vertex '" << name << "' in edge\n"
                                        "  Check spelling and make sure you declared the vertices first.\n";
                                // list verticies?
                                break;
                              }
                            int mult;
                            if (! (istr >> mult)) // get multiplicity
                              mult=1;
                            ret=graph.AddEdge(v1, v2, mult);
                          }
                        if (ret==-1)
                          cerr << "Error adding edge\n";
                        else
                          someoutput=true;
                      }
                      break;

                    case 'c':
                      {
                        incover=true;
                        Coord yscale = graph.Bounds[1][1]+graph.Bounds[1][2]
                                      -graph.Bounds[0][1]-graph.Bounds[0][2]
                                      + setting.DiscRThick + setting.LoopSize;
                        yscale *= setting.CoverYStretch;
                        cover.Initialize(&graph, yscale);
                      }
                      break;
                    default:
                      cerr << "Unrecognized command \"" << s << "\",\n"
                        "  expected 's' 'e' 'v' 'c' or comment beginning with a symbol.\n";
                      needhelp=true;
                      break;
                  }
              }
            else if (s.length())
              instl getline(cin, s);
          }
      }
    return needhelp || ! someoutput;
  }

bool ParseCommandLine(int argc, char *argv[], DrawTo &drawto)
  {
    bool needhelp=false;
    for (int i=1; i<argc; ++i)
      { // process each command line parameter
        string istr(argv[i]);

        int j=istr.find('=');
        if (j==string::npos)
          j=istr.find('#'); // alternate = sign

        if (j==string::npos)
          {
            needhelp=true;
                // check if just crying for help
            if (istr!="/h"    && istr!="-h"    && istr!="h" &&
                istr!="/help" && istr!="-help" && istr!="help" &&
                istr!="/?"    && istr!="-?"    && istr!="?")
              cerr << "Error interpretting parameter " << i << ", \""
                   << argv[i] << "\".\n";
          }
        else
          {
            string name=istr.substr(0, j);
            string valstr=istr.substr(j+1);
            Coord value=atof(valstr.c_str());
            ParseSetting(drawto, name, value);
          }
      }
    return needhelp;
  }

// ---------------------- Program entry point

int main(int argc, char *argv[])
  {
    int i;
    DrawTo drawto;
    Scene scene(&drawto);
    GraphBase graph;
    GraphCover cover;

    bool needhelp=ParseCommandLine(argc, argv, drawto);
    if (! needhelp)
      needhelp=ParseInput(drawto, graph, cover);
    if (! needhelp)
      ParseCommandLine(argc, argv, drawto);

    if (needhelp)
      {
        HelpText();
        return 1;
      }

    graph.UpdateBounds(&drawto);
    Coord xlateby[3];
    for (i=0; i<3; ++i)
      xlateby[i]= -graph.Bounds[i ? 1 : 0][i];
    graph.Translate(xlateby);

    if (cover.WasInitialized())
      {
        cover.Translate(xlateby);
        // optimize covering ...
        cover.Rearrange();
        cover.UpdateBounds(&drawto);
        for (i=0; i<3; ++i)
          xlateby[i]= -cover.Bounds[0][i];
        if (xlateby[0]<0)
          xlateby[0]=0;
        cover.Translate(xlateby);
        xlateby[1]=xlateby[2]=0;
        graph.Translate(xlateby);

        drawto.SetBounds(instl max(graph.Bounds[1][0], cover.Bounds[1][0]),
                         cover.Bounds[1][1]+cover.Bounds[1][2]
                          -graph.Bounds[0][1]-graph.Bounds[0][2]);
        if (up==-1)
          { // deal with the up=+/-1 mess
            for (i=0; i<3; ++i)
              xlateby[i]= i ? -cover.Bounds[1][i] : 0;
          }
        else
          {
            for (i=0; i<3; ++i)
              xlateby[i]= i ? -graph.Bounds[0][i] : 0;
          }
        graph.Translate(xlateby);
        cover.Translate(xlateby);
      }
    else
      {
        drawto.SetBounds(graph.Bounds[1][0],
                        -graph.Bounds[0][1]-graph.Bounds[0][2]);
        if (up!=-1)
          {
            for (i=0; i<3; ++i)
              xlateby[i]= i ? -graph.Bounds[0][i] : 0;
            graph.Translate(xlateby);
          }
      }

    graph.Render(scene);
    if (cover.WasInitialized())
      cover.Render(scene);
    scene.Render();
    drawto.Done();

    // should do a better job of clean up ...

    return 0;
  }
