#include <jstring.h>
#include <vector>
#include <list>
#include "drawgraph.h"
#include "elttypes.h"
#include <assert.h>
#include <jiostr.h>

// ---------------------- Graph member functions

int Graph :: FindVertex(string label)
  {
    int numvert=NumVert();
    for (int i=0; i<numvert; ++i)
      if (Vert(i).label==label)
        return i;
    return -1;
  }

int Graph :: FindVertexSubstring(string label)
  {
    int numvert=NumVert();
    for (int i=0; i<numvert; ++i)
      if (label.find(Vert(i).label) != string::npos)
        return i;
    return -1;
  }

void Graph :: UpdateBounds(DrawTo *drawto)
  {
    int i;
    int numvert=NumVert();
    for (i=0; i<numvert; ++i)
      Vert(i).UpdateBounds(*this, drawto);
    for (i=0; i<e.size(); ++i)
      e[i].UpdateBounds(*this);
  }

void Graph :: Translate(Coord t[3])
  {
    int i;
    int numvert=NumVert();
    for (i=0; i<numvert; ++i)
      Vert(i).Translate(t);
    for (i=0; i<e.size(); ++i)
      e[i].Translate(t);
    for (i=0; i<3; ++i)
      {
        Center[i]+=t[i];
        Bounds[0][i]+=t[i];
        Bounds[1][i]+=t[i];
      }
  }

void Graph :: Render(Scene &scene)
  {
    int i;
    int numvert=NumVert();
    for (i=0; i<numvert; ++i)
      Vert(i).Render(scene);
    for (i=0; i<e.size(); ++i)
      e[i].Render(scene);
  }

int Graph :: AddEdge(int v1, int v2, int multiplicity)
  {
    assert(v1>=0);
    assert(v2>=0);
    assert(v1<NumVert());
    assert(v2<NumVert());
    if (multiplicity<1)
      return -1;
    int i=FindEdge(v1,v2);
    if (i>=0)
      {
        e[i].IncreaseMultiplicity(multiplicity, *this);
        return i;
      }
    e.push_back();
    e.back().Set(v1, v2, multiplicity, *this);
    return e.size()-1;
  }

int Graph :: FindEdge(int v1, int v2)
  {
    assert(v1>=0);
    assert(v2>=0);
    assert(v1<NumVert());
    assert(v2<NumVert());
    for (int i=0; i<e.size(); ++i)
      if ((e[i].v[0]==v1 && e[i].v[1]==v2) ||
          (e[i].v[1]==v1 && e[i].v[0]==v2))
        return i;
    return -1;
  }

// GraphBase -----------------------

Vertex &GraphBase :: Vert(int i)
  {
    return v[i];
  }

int GraphBase :: NumVert()
  {
    return v.size();
  }

int GraphBase :: AddVertex(Coord P[3], string label)
  {
    int i;
    if (v.empty())
      for (i=0; i<3; ++i)
        {
          Bounds[0][i] = P[i];
          Bounds[1][i] = P[i];
        }
    else
      for (i=0; i<3; ++i)
        {
          Bounds[0][i] = instl min(Bounds[0][i], P[i]);
          Bounds[1][i] = instl max(Bounds[1][i], P[i]);
        }

    for (i=0; i<3; ++i)
      Center[i] = (Bounds[0][i]+Bounds[1][i])/2;

    v.push_back();
    v.back().Set(P, label);
    return v.size()-1;
  }

// GraphCover ----------------------

Vertex &GraphCover :: Vert(int i)
  {
    return (Vertex &) v[i];
  }

int GraphCover :: NumVert()
  {
    return v.size();
  }

void GraphCover :: Initialize(GraphBase *newbase, Coord newyspacing)
  { 
    base=newbase; 
    vf.resize(base->NumVert()); 
    ef.resize(base->e.size()); 
    yspacing = newyspacing;
  }

Coord urand()
  {
    Coord ret=rand();
    ret /= RAND_MAX;
    return 2*ret-1;
  }

int GraphCover :: AddVertex(string label)
  {
    int basevert = base->FindVertexSubstring(label);
    if (basevert == -1)
      return -1;

    assert(basevert>=0 && basevert < vf.size());

    // create new vertex
    v.push_back();

    // compute position over original and others in fiber
    Coord newP[3];
    int i;
    for (i=0; i<3; ++i)
      newP[i]=base->Vert(basevert).P[i];
    Coord adjust = urand()*setting.CoverRandomDisplace;
    newP[1] += yspacing * vf[basevert].size() + adjust;
    newP[2] -= adjust;
    v.back().Set(newP, label);
    v.back().basevertex=basevert;

    // add to fiber
    vf[basevert].push_back(v.size()-1);

    // update bounds
    if (v.size()==1)
      for (i=0; i<3; ++i)
        {
          Bounds[0][i] = newP[i];
          Bounds[1][i] = newP[i];
        }
    else
      for (i=0; i<3; ++i)
        {
          Bounds[0][i] = instl min(Bounds[0][i], newP[i]);
          Bounds[1][i] = instl max(Bounds[1][i], newP[i]);
        }

    // update center
    for (i=0; i<3; ++i)
      Center[i] = (Bounds[0][i]+Bounds[1][i])/2;

    return v.size()-1;
  }

void GraphCover :: Rearrange()
  {
    // for each base edge
    if (setting.CoverEnableCurves)
      for (int baseedge=0; baseedge<base->e.size(); ++baseedge)
        {
          // if multiplicity > 1 or v[0]==v[1]
          if (base->e[baseedge].arcs.size()>1 || 
              base->e[baseedge].v[0]==base->e[baseedge].v[1])
            {
              bool ok=AssignCoverEdge(this, baseedge);
              // report an error if necessary
              if (! ok)
                cerr << "Error finding covering for base edge " << baseedge << ".";
            }
          // otherwise it should be fine
        }
    else
      for (int i=0; i<e.size(); ++i)
        {
          int multiplicity=e[i].arcs.size();
          e[i].arcs.clear();
          e[i].Set(e[i].v[0], e[i].v[1], multiplicity, *this);
        }
  }

int GraphCover :: AddEdge(int v1, int v2, int multiplicity)
  {
    assert(v1>=0);
    assert(v2>=0);
    assert(v1<NumVert());
    assert(v2<NumVert());
    int baseedge=base->FindEdge(v[v1].basevertex, v[v2].basevertex);
    if (baseedge<0)
      return -1;
    int ret=Graph::AddEdge(v1, v2, multiplicity);
    if (e[ret].arcs.size()>multiplicity)
      { // added multiplicity to existing edge
//        ef[baseedge].mV[v1].DegreeLeft += multiplicity;
//        ef[baseedge].mV[v2].DegreeLeft += multiplicity;
      }
        // new edge
    else if (v1==v2)  
      { // new loop
//        ef[baseedge].mV[v1].DegreeLeft += multiplicity*2;
        ef[baseedge].mV[v1].Edges.push_back(EdgeFiber::PerEdge(v1, ret));
      }
    else // (v1!=v2)
      {
//        ef[baseedge].mV[v1].DegreeLeft += multiplicity;
//        ef[baseedge].mV[v2].DegreeLeft += multiplicity;
        ef[baseedge].mV[v1].Edges.push_back(EdgeFiber::PerEdge(v2, ret));
        ef[baseedge].mV[v2].Edges.push_back(EdgeFiber::PerEdge(v1, ret));
      }
    return ret;
  }
