#include <jstring.h>
#include <vector>
#include <list>
#include "drawgraph.h"
#include <set>
#include <assert.h>
#include <algorithm>

namespace {

class ConstrainedSearchSpace
  {
public:
    typedef ConstrainedSearchSpace self;

    // Required:
    bool SetCurVariableMostConstrained(); // False -> done
    bool AdvanceCurVariable();
    bool SetCurValueLeastConstraining();
    bool AdvanceCurValue();
    bool AssignCurrent();                 // specific. False -> led to an inconsistancy
    bool success;                         // True -> everything assigned, set by AssignCurrent
                                          // initialized to false
    // if default not good: self(const self &); // is this a constructor?
    // if default not good: self & operator=(const self &);
    
    // default implimentation:
    typedef unsigned variabletype;           // specific
    typedef unsigned rangetype;              // specific

    struct varvaluepair
      {
        variabletype var;
        rangetype    value;
      };

    // This data structure is a bit too general for the n-queens problem since
    // you can get by with a single possible list, but this is good for the
    // graph numbering problem I'm concerned about, as well as other hard problems

    typedef instl set<rangetype>   possiblelist;
    typedef possiblelist::iterator possibleiter;

    typedef instl map<variabletype, possiblelist> varlisttype;
    typedef varlisttype::iterator                 varlistiter;
    typedef instl pair<const variabletype, possiblelist> varlistitem;

    varlisttype UnassignedVars;

    static ConstrainedSearchSpace *HACKcurrent;

    struct comparevarpossibilities
      {
        bool operator() (variabletype v1, variabletype v2)
          {
            varlistiter c1=HACKcurrent->UnassignedVars.find(v1);
            varlistiter c2=HACKcurrent->UnassignedVars.find(v2);
            assert(c1!=HACKcurrent->UnassignedVars.end());
            assert(c2!=HACKcurrent->UnassignedVars.end());
            return (*c1).second.size()<(*c2).second.size() ||
                    ((*c1).second.size()==(*c2).second.size() && 
                    v1<v2);
          }
      };

    typedef instl set<variabletype, comparevarpossibilities> varordertype;
    typedef varordertype::iterator                           varorderiter;

    varordertype VariablePriorities;  // MUST SET HACKcurrent ANYTIME YOU MANIPULATE OR EVEN SEARCH
    varorderiter CurrentVariable;
    
    struct valueconstrainingpair
      {
        typedef valueconstrainingpair self;
        rangetype value;
        int numoptionsremoved;

        valueconstrainingpair(rangetype newvalue, int newnumoptionsremoved)
          { value=newvalue; numoptionsremoved=newnumoptionsremoved; }
        inline bool operator<(const self &c2) const
          { 
            const self &c1=*this;
            return c1.numoptionsremoved<c2.numoptionsremoved ||
                   (c1.numoptionsremoved==c2.numoptionsremoved && 
                    c1.value<c2.value);
          }
      };
    typedef instl vector<valueconstrainingpair> valuelisttype;
    typedef valuelisttype::iterator             valuelistiter;
    valuelisttype PossibleValues;
    valuelistiter CurrentValue;

    int GetNumOptionsRemovedBy(rangetype value);  // specific.  uses CurrentVariable

/*
Specifics to arranging a GraphCover

Fixed for the duration of the search: 
  An instance of GraphCover (mCover) and a base edge (mEdge)

Input: one base edge with multiplicity at least 2 
    or one base loop (any multiplicity).  
Result: assign a base edge for each covering edge

typedef unsigned variabletype;
  variabletype e;
  (e >> edgenumshift) & edgenummask   -- gives which edge number, an index into GraphCover::e
  (e >> subedgeshift) & subedgemask   -- index within multiedges

typedef unsigned rangetype;
  rangetype r;
  (r >> subedgeshift) & subedgemask   -- gives index within list of edges in base 
  (r >> reversedshift) & reversedmask -- in the case of loops, indicates if the edge should go backwards

Note that GraphCover::e stores the the two incident cover vertices for a given cover (multi)edge
Given a cover vertex v, 
  GraphCover::ef[mBaseEdge].mV[v]     -- has the cover edges incident to that vertex over the base edge
  GraphCover::v[v].basevertex         -- has the base vertex

GraphCover::base->e[mBaseEdge].v[] has the two vertices incident to the base edge
GraphCover::base->e[mBaseEdge].arcs.size() has number of edges in base multiedge
*/
    GraphCover *mCover;
    int         mBaseEdge;

    typedef instl map<variabletype, rangetype> AssignedValuesType;
    typedef AssignedValuesType::iterator       AssignedValuesIter;
    AssignedValuesType mAssignedValues;

    enum { subedgeshift = 0, edgenumshift = 4*sizeof(unsigned), 
           subedgemask  = ((1<<edgenumshift)-1), edgenummask=subedgemask,
           reversedshift = edgenumshift, reversedmask = 1 };
  };

ConstrainedSearchSpace * ConstrainedSearchSpace :: HACKcurrent=0;

typedef ConstrainedSearchSpace :: valueconstrainingpair vcp;
/*
inline bool operator<(const vcp &c1, const vcp &c2)
  { 
    return c1.numoptionsremoved<c2.numoptionsremoved ||
           (c1.numoptionsremoved==c2.numoptionsremoved && 
            c1.value<c2.value);
  }
*/

bool ConstrainedSearchSpace :: SetCurVariableMostConstrained()
  {
    HACKcurrent=this;
    CurrentVariable=VariablePriorities.begin();
    return CurrentVariable!=VariablePriorities.end();
  }

bool ConstrainedSearchSpace :: AdvanceCurVariable()
  {
    HACKcurrent=this;
    assert(CurrentVariable!=VariablePriorities.end());
    ++CurrentVariable;
    return CurrentVariable!=VariablePriorities.end();
  }

bool ConstrainedSearchSpace :: SetCurValueLeastConstraining()
  {
    PossibleValues.clear();
    HACKcurrent=this;
    for (possibleiter p=UnassignedVars[(*CurrentVariable)].begin();
         p != UnassignedVars[(*CurrentVariable)].end();
         ++p)
      PossibleValues.push_back(valueconstrainingpair(*p, GetNumOptionsRemovedBy(*p)));
    instl sort(PossibleValues.begin(), PossibleValues.end(), instl less<vcp>());
    CurrentValue=PossibleValues.begin();
    return CurrentValue!=PossibleValues.end();
  }

bool ConstrainedSearchSpace :: AdvanceCurValue()
  {
    assert(CurrentValue!=PossibleValues.end());
    ++CurrentValue;
    return CurrentValue!=PossibleValues.end();
  }

// THE ALGORITHM

bool ConstrainedSearch(ConstrainedSearchSpace &search)
  {
    bool ok = search.SetCurVariableMostConstrained();
    while (ok)
      {
        ok = search.SetCurValueLeastConstraining();
        while (ok)
          {
            ConstrainedSearchSpace cpy(search);
            if (cpy.AssignCurrent())
              if (cpy.success || ConstrainedSearch(cpy))
                { // Found a solution!
                  search = cpy;
                  return true;
                }
            ok = search.AdvanceCurValue();
          }
        ok=search.AdvanceCurVariable();
      }
    return false;
  }


// SPECIFICS

// uses CurrentVariable
int ConstrainedSearchSpace :: GetNumOptionsRemovedBy(rangetype value)
  {
    HACKcurrent=this;
    variabletype var=(*CurrentVariable);
    int ret=0;
        // for every variable which would be affected by this assignment
    for (int whichv=0; whichv<2; ++whichv)
      {
        int cv=mCover->e[(var >> edgenumshift) & edgenummask].v[whichv];
        EdgeFiber::edgelist *edges=& mCover->ef[mBaseEdge].mV[cv].Edges;
        EdgeFiber::edgeiter ce=edges->begin();
        rangetype excludedpossibility=value;
        if (whichv==1 && /* loop in base: */
            mCover->base->e[mBaseEdge].v[0]==mCover->base->e[mBaseEdge].v[1])
          excludedpossibility ^= (reversedmask << reversedshift);
        for (; ce!=edges->end(); ++ce)
          {
            int edgenum=(*ce).edgenum;
            for (int subedge=0; subedge < mCover->e[edgenum].arcs.size(); ++subedge)
              {
                variabletype othervar=(edgenum << edgenumshift) + (subedge << subedgeshift);
                if (othervar!=var)
                  { // othervar will take on all edges sharing a vertex with var
                    varlistiter isunassigned=UnassignedVars.find(othervar);
                    // check make sure not already assigned
                    if (isunassigned!=UnassignedVars.end())
                      {
                        // figure out if this is a possibility for othervar
                        rangetype actuallyexcluded=excludedpossibility;
                        if (mCover->e[edgenum].v[0]!=cv && /* loop in base: */
                              mCover->base->e[mBaseEdge].v[0]==mCover->base->e[mBaseEdge].v[1])
                          actuallyexcluded ^= (reversedmask << reversedshift);

                        possibleiter possiblevalue=(*isunassigned).second.find(actuallyexcluded);
                        if (possiblevalue!=(*isunassigned).second.end())
                          ++ret;  // it would be an option removed by the proposed assignment
                      }
                  }
              }
          }
      }

    return 0;
  }

// returning False means assigning the current led to an inconsistancy
bool ConstrainedSearchSpace :: AssignCurrent()
  {
    // List of values to assign
    instl list<varvaluepair> toassign;
    varvaluepair toadd={(*CurrentVariable), (*CurrentValue).value};
    toassign.push_back(toadd);
    
    HACKcurrent=this;
    CurrentVariable=VariablePriorities.end(); // prevent CurrentVariable from being used

    // while list not empty
    for (; ! toassign.empty(); toassign.erase(toassign.begin()))
      {
        // var = variable to assign
        variabletype var=toassign.front().var;
        // value = value to assign to var
        rangetype    value=toassign.front().value;
        // get pointer to possible values for var
        varlistiter  beforeassign=UnassignedVars.find(var);
        // check make sure not already assigned
        if (beforeassign==UnassignedVars.end())
          {
            assert(mAssignedValues[var]==value);
            continue;
          }
        // check make sure assigning one of the possible values
        possibleiter ispossible=(*beforeassign).second.find(value);
        if (ispossible==(*beforeassign).second.end())
          return false;  // Bad: trying to assign a value not among the possibilities

        // assign var
        mAssignedValues[var]=value;
        // remove var from list of unassigned
        VariablePriorities.erase(var);
        UnassignedVars.erase(beforeassign);

        // for every variable affected by this assignment
        for (int whichv=0; whichv<2; ++whichv)
          {
            int cv=mCover->e[(var >> edgenumshift) & edgenummask].v[whichv];
            EdgeFiber::edgelist *edges=& mCover->ef[mBaseEdge].mV[cv].Edges;
            EdgeFiber::edgeiter ce=edges->begin();
            rangetype excludedpossibility=value;
            if (whichv==1 && /* loop in base: */
                mCover->base->e[mBaseEdge].v[0]==mCover->base->e[mBaseEdge].v[1])
              excludedpossibility ^= (reversedmask << reversedshift);
            for (; ce!=edges->end(); ++ce)
              {
                int edgenum=(*ce).edgenum;
                for (int subedge=0; subedge < mCover->e[edgenum].arcs.size(); ++subedge)
                  {
                    variabletype othervar=(edgenum << edgenumshift) + (subedge << subedgeshift);
                    if (othervar!=var)
                      { // othervar will take on all edges sharing a vertex with var
                        varlistiter isunassigned=UnassignedVars.find(othervar);
                        // check make sure not already assigned
                        if (isunassigned!=UnassignedVars.end())
                          {
                            // Can alter possibility list directly, but then must update VariablePriorities
                            VariablePriorities.erase(othervar);

                            rangetype actuallyexcluded=excludedpossibility;
                            if (mCover->e[edgenum].v[0]!=cv && /* loop in base: */
                                mCover->base->e[mBaseEdge].v[0]==mCover->base->e[mBaseEdge].v[1])
                              actuallyexcluded ^= (reversedmask << reversedshift);

                            // figure out updated possibility list
                            (*isunassigned).second.erase(actuallyexcluded);

                            // if it has no possibilities now
                            if ((*isunassigned).second.empty())
                              return false; // 
                            else /* if it only has one possibility now */
                              if ((*isunassigned).second.size()==1)
                                { // add it to the list of variables to assign
                                  toadd.var=othervar;
                                  toadd.value=(*(*isunassigned).second.begin());
                                  toassign.push_back(toadd);
                                }
                            else
                                // update that variable's possible list 
                                // by re-adding it to VariablePriorities
                              VariablePriorities.insert(othervar);
                          }
                      }
                  }
              }
          }
      }
    
    success=(UnassignedVars.empty());
    return true;
  }

} /* end of anonymous namespace */

bool AssignCoverEdge(GraphCover *cover, int baseedge)
  {
    ConstrainedSearchSpace search;
    
    { // initialize search 
      search.mCover = cover;
      search.mBaseEdge = baseedge;
      ConstrainedSearchSpace :: HACKcurrent = &search;

      // create a possible list of either:
      //    all subedges of a multiedge in the base, or
      //    all subloops and reverses of those subloops of a (multi)loop in the base
      ConstrainedSearchSpace :: possiblelist possible;
      if (cover->base->e[baseedge].v[0] == cover->base->e[baseedge].v[1])
        { // base is a loop, generate reverses as well
          for (int e=0; e<cover->base->e[baseedge].arcs.size(); ++e)
            for (int r=0; r<2; ++r)
              possible.insert((e<<ConstrainedSearchSpace::subedgeshift) + 
                              (r<<ConstrainedSearchSpace::reversedshift));
        }
      else
        { // base is an edge between distinct vertices
          for (int e=0; e<cover->base->e[baseedge].arcs.size(); ++e)
            possible.insert(e<<ConstrainedSearchSpace::subedgeshift);
        }

      // add all edge over baseedge to search
      for (unsigned cvi=0; cvi < cover->vf[cover->base->e[baseedge].v[0]].size(); ++cvi)
        {
          int cv=cover->vf[cover->base->e[baseedge].v[0]][cvi];
          for (EdgeFiber :: edgeiter edge = cover->ef[baseedge].mV[cv].Edges.begin();
               edge != cover->ef[baseedge].mV[cv].Edges.end();
               ++edge)
            {
              // see if already added this edge (an issue for loop covers)
              unsigned edgenum = (*edge).edgenum;
              ConstrainedSearchSpace::variabletype edgename = 
                    edgenum << ConstrainedSearchSpace::edgenumshift;

              ConstrainedSearchSpace::varlistiter esofar;
              esofar=search.UnassignedVars.find(edgename);

              if (esofar==search.UnassignedVars.end())
                { // not already present in UnassignedVars, so add
                  for (int subedge = 0; subedge < cover->e[edgenum].arcs.size(); ++subedge)
                    {
                      ConstrainedSearchSpace::variabletype var 
                          = edgename + (subedge << ConstrainedSearchSpace::subedgeshift);
                      search.UnassignedVars.insert(ConstrainedSearchSpace::varlistitem(var, possible));
                      search.VariablePriorities.insert(var);
                    }
                }
            }
        }
    }

    bool result = ConstrainedSearch(search);
    if (result)
      { // repath each edge according to search results
        for (ConstrainedSearchSpace::AssignedValuesIter cei=search.mAssignedValues.begin();
             cei != search.mAssignedValues.end(); ++ cei)
          {
            // (*cei).first is const variabletype, (*cei).second is rangetype
            int coveredge = ((*cei).first >> ConstrainedSearchSpace::edgenumshift) 
                            & ConstrainedSearchSpace::edgenummask;
            int coversubedgenum = ((*cei).first >> ConstrainedSearchSpace::subedgeshift) 
                            & ConstrainedSearchSpace::subedgemask;
            int basesubedgenum = ((*cei).second >> ConstrainedSearchSpace::subedgeshift) 
                            & ConstrainedSearchSpace::subedgemask;
            bool reversed = ((*cei).second >> ConstrainedSearchSpace::reversedshift) 
                            & ConstrainedSearchSpace::reversedmask;
            // find base subedge
            assert(basesubedgenum<cover->base->e[baseedge].arcs.size());
            Edge::arciter basesubedge= cover->base->e[baseedge].arcs.begin();
            while (basesubedgenum--)
               ++basesubedge;
            assert(basesubedge != cover->base->e[baseedge].arcs.end());

            // find cover subedge
            assert(coversubedgenum<cover->e[coveredge].arcs.size());
            Edge::arciter coversubedge= cover->e[coveredge].arcs.begin();
            while (coversubedgenum--)
               ++coversubedge;
            assert(coversubedge != cover->e[coveredge].arcs.end());

            // base vertices
            int b1=cover->base->e[baseedge].v[0];
            int b2=cover->base->e[baseedge].v[1];

            // cover vertices
            int c1=cover->e[coveredge].v[0];
            int c2=cover->e[coveredge].v[1];

            // if reversed, swap cover vertices 
            //   (or do we actually have to swap base vertices and iterate backwards?)
            if (b1==b2 && reversed)
              instl swap(c1, c2);

            // correction = delta cover vertices minus delta base vertices
            Coord correctionY = cover->v[c2].P[1] - cover->v[c1].P[1] - 
                                (cover->base->v[b2].P[1] - cover->base->v[b1].P[1]);
            Coord correctionZ = cover->v[c2].P[2] - cover->v[c1].P[2] - 
                                (cover->base->v[b2].P[2] - cover->base->v[b1].P[2]);
            // offset = first cover vertex - first base vertex
            Coord offsetY = cover->v[c1].P[1] - cover->base->v[b1].P[1];
            Coord offsetZ = cover->v[c1].P[2] - cover->base->v[b1].P[2];
            // clear coversubedge
            (*coversubedge).clear();
            // iterate through basesubedge splinepoints
            Coord t=0;
            Coord dt=1./((*basesubedge).size()-1);
            for (EdgeSplineIter splinepoint=(*basesubedge).begin();
                  splinepoint!=(*basesubedge).end();
                  ++splinepoint, t+=dt)
              {
                // add coversubedge splinepoints according to:
                //  basesplinepoint + t * correction + offset
                GraphPoint toadd = *splinepoint;
                toadd.p[1] += t*correctionY + offsetY;
                toadd.p[2] += t*correctionZ + offsetZ;
                (*coversubedge).push_back(toadd);
              }

            // do we need to adjust neighboring vertices?
            cover->e[coveredge].ModifyNeighbors(*cover);
          }
      }
    return result;
  }

  
